"""CLI interface for yen."""

from __future__ import annotations

import argparse
import os.path
import subprocess
import sys
from typing import Literal

from yen import (
    DEFAULT_PYTHON_VERSION,
    PACKAGE_INSTALLS_PATH,
    ExecutableDoesNotExist,
    check_path,
    create_venv,
    ensure_python,
    ensurepath,
    install_package,
)
from yen.github import NotAvailable, list_pythons


class YenArgs:
    command: Literal["list", "ensurepath", "create", "install", "run", "exec"]
    python: str
    venv_path: str
    package_name: str
    binary: str | None
    module: str | None
    force_reinstall: bool
    run_args: list[str]


def cli() -> int:
    """CLI interface."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command", required=True)

    subparsers.add_parser("list")
    subparsers.add_parser("ensurepath")

    create_parser = subparsers.add_parser("create")
    create_parser.add_argument("venv_path", type=os.path.abspath)
    create_parser.add_argument("-p", "--python", required=True)

    install_parser = subparsers.add_parser("install")
    install_parser.add_argument("package_name")
    install_parser.add_argument("-p", "--python", default=DEFAULT_PYTHON_VERSION)
    install_parser.add_argument(
        "--binary",
        help="Name of command installed by package. Defaults to package name itself.",
    )
    install_parser.add_argument(
        "--module",
        help="Use if package should be run as a module, i.e. `python -m <module_name>`",
    )
    install_parser.add_argument("--force-reinstall", action="store_true")

    # TODO: add long help texts to each subparser
    run_parser = subparsers.add_parser("run")
    run_parser.add_argument("package_name")
    run_parser.add_argument("-p", "--python", default=DEFAULT_PYTHON_VERSION)
    run_parser.add_argument(
        "run_args",
        help="Arguments to pass to the command invocation",
        nargs="*",
    )

    exec_parser = subparsers.add_parser("exec")
    exec_parser.add_argument("-p", "--python", default=DEFAULT_PYTHON_VERSION)

    args = parser.parse_args(namespace=YenArgs)

    if args.command == "list":
        versions = list(list_pythons())
        print("Available Pythons:", file=sys.stderr)
        for version in versions:
            print(version)

    if args.command == "ensurepath":
        ensurepath()
        print(
            f"`{PACKAGE_INSTALLS_PATH}` is now present in your PATH."
            " Restart your shell for it to take effect."
        )

    elif args.command == "create":
        try:
            python_version, python_bin_path = ensure_python(args.python)
        except NotAvailable:
            print(
                "Error: requested Python version is not available."
                " Use 'yen list' to get list of available Pythons.",
                file=sys.stderr,
            )
            return 1

        if os.path.exists(args.venv_path):
            print(f"\033[1;31mError:\033[m {args.venv_path} already exists.")
            return 2

        create_venv(python_bin_path, args.venv_path)
        print(f"Created \033[1m{args.venv_path}\033[m with Python {python_version} ✨")

    elif args.command == "install":
        if args.module is not None and args.binary is not None:
            print(
                "Error: cannot pass `--binary` and `--module` together.",
                file=sys.stderr,
            )
            return 1

        try:
            python_version, python_bin_path = ensure_python(args.python)
        except NotAvailable:
            print(
                "Error: requested Python version is not available."
                " Use 'yen list' to get list of available Pythons.",
                file=sys.stderr,
            )
            return 1

        # TODO: add yaspin?
        executable_name = args.module or args.binary or args.package_name
        is_module = args.module is not None
        try:
            _, already_installed = install_package(
                args.package_name,
                python_bin_path,
                executable_name,
                is_module=is_module,
                force_reinstall=args.force_reinstall,
            )
        except ExecutableDoesNotExist:
            error_message = (
                f"Error: package {args.package_name} doesn't contain a binary named"
                f" {executable_name}."
            )
            if not (args.module or args.binary):
                error_message += " Consider passing `--binary` or `--module` flags."
            print(error_message, file=sys.stderr)
            return 4

        if already_installed:
            print(f"Package \033[1m{args.package_name}\033[m is already installed.")
        else:
            print(
                f"Installed package \033[1m{args.package_name}\033[m"
                f" with Python {python_version} ✨"
            )

        check_path(PACKAGE_INSTALLS_PATH)

    elif args.command == "run":
        try:
            _, python_bin_path = ensure_python(args.python)
        except NotAvailable:
            print(
                "Error: requested Python version is not available."
                " Use 'yen list' to get list of available Pythons.",
                file=sys.stderr,
            )
            return 1

        # TODO: add yaspin?
        try:
            shim_path, _ = install_package(
                args.package_name,
                python_bin_path,
                executable_name=args.package_name,
            )
        except ExecutableDoesNotExist:
            print(
                f"Error: package {args.package_name} doesn't contain a binary named"
                f" {args.package_name}.",
                file=sys.stderr,
            )
            return 4

        return subprocess.call([shim_path, *args.run_args])

    elif args.command == "exec":
        try:
            python_version, python_bin_path = ensure_python(args.python)
        except NotAvailable:
            print(
                "Error: requested Python version is not available."
                " Use 'yen list' to get list of available Pythons.",
                file=sys.stderr,
            )
            return 1

        return subprocess.call([python_bin_path])

    return 0



"""Taken from https://github.com/textualize/rich/blob/ec91917/examples/downloader.py"""

from __future__ import annotations

from http.client import HTTPResponse
import os.path
import signal
from functools import partial
from threading import Event
from urllib.request import urlopen

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)


PROGRESS = Progress(
    TextColumn("[bold blue]{task.fields[display_name]}"),
    BarColumn(bar_width=None),
    "[progress.percentage]{task.percentage:>3.1f}%",
    "•",
    DownloadColumn(),
    "•",
    TransferSpeedColumn(),
    "•",
    TimeRemainingColumn(),
)


DONE = Event()


def handle_sigint(_: object, __: object) -> None:
    DONE.set()


signal.signal(signal.SIGINT, handle_sigint)


def read_url(url: str) -> str:
    """Reads the contents of the URL."""
    response: HTTPResponse = urlopen(url)
    return response.read().decode()


def download(url: str, display_name: str, directory: str) -> str:
    """Downloads file to the given directory. Returns path to downloaded file."""
    with PROGRESS:
        filename = url.split("/")[-1]
        filepath = os.path.join(directory, filename)
        task_id = PROGRESS.add_task("download", display_name=display_name, start=False)
        response = urlopen(url)

        # This will break if the response doesn't contain content length
        PROGRESS.update(task_id, total=int(response.info()["Content-length"]))
        with open(filepath, "wb") as file:
            PROGRESS.start_task(task_id)
            for data in iter(partial(response.read, 32768), b""):
                file.write(data)
                PROGRESS.update(task_id, advance=len(data))

    return filepath



"""yen - Yet another Python environment manager."""

from __future__ import annotations

import hashlib
import os
import os.path
import platform
import shutil
import subprocess
import sys
import tarfile
from urllib.request import urlretrieve

from yen.downloader import download, read_url
from yen.github import resolve_python_version

YEN_BIN_PATH = os.path.abspath(
    os.getenv("YEN_BIN_PATH", os.path.expanduser("~/.yen/bin"))
)
PYTHON_INSTALLS_PATH = os.path.abspath(
    os.getenv("YEN_PYTHONS_PATH", os.path.expanduser("~/.yen_pythons"))
)
PACKAGE_INSTALLS_PATH = os.path.abspath(
    os.getenv("YEN_PACKAGES_PATH", os.path.expanduser("~/.yen_packages"))
)

USERPATH_PATH = os.path.join(YEN_BIN_PATH, "userpath.pyz")
MICROVENV_PATH = os.path.join(YEN_BIN_PATH, "microvenv.py")

DEFAULT_PYTHON_VERSION = "3.12"


class ExecutableDoesNotExist(Exception): ...


def check_path(path: str) -> None:
    """Check if given path is in PATH, and inform the user otherwise."""
    if platform.system() == "Windows":
        _ensure_userpath()
        python_bin_path = find_or_download_python()
        process = subprocess.run([python_bin_path, USERPATH_PATH, "check", path])
        path_exists = process.returncode == 0
    else:
        path_exists = path in os.environ["PATH"].split(os.pathsep)

    if not path_exists:
        print(
            "\033[33m"
            "Warning: The executable just installed is not in PATH.\n"
            "Run `yen ensurepath` to add it to your PATH."
            "\033[m",
            file=sys.stderr,
        )


def _ensure_userpath() -> None:
    """Downloads `userpath.pyz`, if it doesn't exist in `YEN_BIN_PATH`."""
    if os.path.exists(USERPATH_PATH):
        return

    os.makedirs(YEN_BIN_PATH, exist_ok=True)
    urlretrieve("http://yen.tushar.lol/userpath.pyz", filename=USERPATH_PATH)


def _ensure_microvenv() -> None:
    """Downloads `microvenv.py`, if it doesn't exist in `YEN_BIN_PATH`."""
    if os.path.exists(MICROVENV_PATH):
        return

    os.makedirs(YEN_BIN_PATH, exist_ok=True)
    urlretrieve("http://yen.tushar.lol/microvenv.py", filename=MICROVENV_PATH)


def find_or_download_python() -> str:
    """
    Finds and returns any Python binary from `PYTHON_INSTALLS_PATH`.
    If no Pythons exist, downloads the default version and returns that.
    """
    for python_folder_name in os.listdir(PYTHON_INSTALLS_PATH):
        python_folder = os.path.join(PYTHON_INSTALLS_PATH, python_folder_name)
        python_bin_path = _python_bin_path(python_folder)
        if os.path.isfile(python_bin_path):
            return python_bin_path

    # No Python binary found. Download one.
    _, python_bin_path = ensure_python(DEFAULT_PYTHON_VERSION)
    return python_bin_path


def ensurepath() -> None:
    """Ensures that PACKAGE_INSTALLS_PATH is in PATH."""
    _ensure_userpath()
    python_bin_path = find_or_download_python()
    subprocess.run(
        [python_bin_path, USERPATH_PATH, "append", PACKAGE_INSTALLS_PATH],
        check=True,
    )


def _python_bin_path(python_directory: str) -> str:
    """Return the python binary path in a downloaded and extracted Python."""
    if platform.system() == "Windows":
        return os.path.join(python_directory, "python", "python.exe")
    else:
        return os.path.join(python_directory, "python", "bin", "python3")


def ensure_python(python_version: str) -> tuple[str, str]:
    """Checks if given Python version exists locally. If not, downloads it."""
    os.makedirs(PYTHON_INSTALLS_PATH, exist_ok=True)

    for python_folder_name in os.listdir(PYTHON_INSTALLS_PATH):
        python_folder = os.path.join(PYTHON_INSTALLS_PATH, python_folder_name)
        if python_folder_name.startswith(python_version):
            # already installed
            python_bin_path = _python_bin_path(python_folder)
            return python_folder_name, python_bin_path

    python_version, download_link = resolve_python_version(python_version)
    download_directory = os.path.join(PYTHON_INSTALLS_PATH, python_version)

    os.makedirs(download_directory, exist_ok=True)
    downloaded_filepath = download(
        download_link,
        f"Downloading {python_version}",
        download_directory,
    )
    # Calculate checksum
    with open(downloaded_filepath, "rb") as python_zip:
        checksum = hashlib.sha256(python_zip.read()).hexdigest()

    # Validate checksum
    checksum_link = download_link + ".sha256"
    expected_checksum = read_url(checksum_link).rstrip("\n")
    if checksum != expected_checksum:
        print("\033[1;31mError:\033[m Checksum did not match!")
        os.remove(downloaded_filepath)
        raise SystemExit(1)
    print("Checksum verified!")

    with tarfile.open(downloaded_filepath, mode="r:gz") as tar:
        tar.extractall(download_directory)

    os.remove(downloaded_filepath)

    python_bin_path = _python_bin_path(download_directory)
    assert os.path.exists(python_bin_path)
    return python_version, python_bin_path


def create_venv(python_bin_path: str, venv_path: str) -> None:
    # if platform.system() == "Windows":
    subprocess.run([python_bin_path, "-m", "venv", venv_path], check=True)
    return

    # _ensure_microvenv()
    # subprocess.run([python_bin_path, MICROVENV_PATH, venv_path], check=True)
    # venv_python_path = _venv_binary_path("python", venv_path)
    # subprocess.run(
    #     [venv_python_path, "-m", "ensurepip"],
    #     check=True,
    #     capture_output=True,
    # )


def _venv_binary_path(binary_name: str, venv_path: str) -> str:
    is_windows = platform.system() == "Windows"
    venv_bin_path = os.path.join(venv_path, "Scripts" if is_windows else "bin")
    binary_path = os.path.join(
        venv_bin_path, f"{binary_name}.exe" if is_windows else binary_name
    )
    return binary_path


def install_package(
    package_name: str,
    python_bin_path: str,
    executable_name: str,
    *,
    is_module: bool = False,
    force_reinstall: bool = False,
) -> tuple[str, bool]:
    is_windows = platform.system() == "Windows"
    shim_path = os.path.join(PACKAGE_INSTALLS_PATH, package_name)
    if is_windows:
        # This is somewhat of a hack.
        # For the condition where shim_path exists and we do `yen run`,
        # `is_module` is false but we still want to return early.
        # But for the condition where we try to create the module the first time,
        # `is_module` will be true, and in that case we want to use `.bat` as well
        if is_module or os.path.exists(shim_path + ".bat"):
            shim_path += ".bat"
        else:
            shim_path += ".exe"

    venv_name = f"venv_{package_name}"
    venv_path = os.path.join(PACKAGE_INSTALLS_PATH, venv_name)
    if os.path.exists(shim_path):
        if not force_reinstall:
            return shim_path, True  # True as in package already existed
        else:
            os.remove(shim_path)
            shutil.rmtree(venv_path, ignore_errors=True)

    create_venv(python_bin_path, venv_path)

    venv_python_path = _venv_binary_path("python", venv_path)
    subprocess.run(
        [venv_python_path, "-m", "pip", "install", package_name],
        check=True,
        capture_output=True,
    )

    if is_module:
        with open(shim_path, "w") as file:
            if is_windows:
                file.write(f"@echo off\n{venv_python_path} -m {package_name} %*")
            else:
                file.write(f'#!/bin/sh\n{venv_python_path} -m {package_name} "$@"')

        os.chmod(shim_path, 0o777)
    else:
        executable_path = _venv_binary_path(executable_name, venv_path)
        if not os.path.exists(executable_path):
            # cleanup the venv created
            shutil.rmtree(venv_path)
            raise ExecutableDoesNotExist

        # the created binary is always moveable
        shutil.move(executable_path, shim_path)

    return shim_path, False  # False as in package didn't exist and was just installed



from __future__ import annotations

import json
import os.path
import platform
import re
import sys
import typing
import urllib.error
from typing import Any, TypedDict
import urllib.parse
from urllib.request import urlopen

LAST_TAG_FOR_I686_LINUX = "118809599"  # tag name: "20230826"

MACHINE_SUFFIX: dict[str, dict[str, Any]] = {
    "Darwin": {
        "arm64": ["aarch64-apple-darwin-install_only.tar.gz"],
        "x86_64": ["x86_64-apple-darwin-install_only.tar.gz"],
    },
    "Linux": {
        "aarch64": {
            "glibc": ["aarch64-unknown-linux-gnu-install_only.tar.gz"],
            # musl doesn't exist
        },
        "x86_64": {
            "glibc": [
                "x86_64_v3-unknown-linux-gnu-install_only.tar.gz",
                "x86_64-unknown-linux-gnu-install_only.tar.gz",
            ],
            "musl": ["x86_64_v3-unknown-linux-musl-install_only.tar.gz"],
        },
        "i686": {
            "glibc": ["i686-unknown-linux-gnu-install_only.tar.gz"],
            # musl doesn't exist
        },
    },
    "Windows": {
        "AMD64": ["x86_64-pc-windows-msvc-install_only.tar.gz"],
        "i686": ["i686-pc-windows-msvc-install_only.tar.gz"],
    },
}

GITHUB_API_RELEASES_URL = (
    "https://api.github.com/repos/astral-sh/python-build-standalone/releases/"
)
PYTHON_VERSION_REGEX = re.compile(r"cpython-(\d+\.\d+\.\d+)")


class GitHubReleaseData(TypedDict):
    id: int
    html_url: str
    assets: list[GitHubAsset]


class GitHubAsset(TypedDict):
    browser_download_url: str


def trim_github_release_data(release_data: dict[str, Any]) -> GitHubReleaseData:
    return {
        "id": release_data["id"],
        "html_url": release_data["html_url"],
        "assets": [
            {"browser_download_url": asset["browser_download_url"]}
            for asset in release_data["assets"]
        ],
    }


def fallback_release_data() -> GitHubReleaseData:
    """Returns the fallback release data, for when GitHub API gives an error."""
    print(
        "\033[33mWarning: GitHub unreachable. Using fallback release data.\033[m",
        file=sys.stderr,
    )
    data_file = os.path.join(os.path.dirname(__file__), "fallback_release_data.json")
    with open(data_file) as data:
        return typing.cast(GitHubReleaseData, json.load(data))


class NotAvailable(Exception):
    """Raised when the asked Python version is not available."""


def get_latest_python_releases(is_linux_i686: bool) -> GitHubReleaseData:
    """Returns the list of python download links from the latest github release."""
    # They stopped shipping for 32 bit linux since after the 20230826 tag
    if is_linux_i686:
        data_file = os.path.join(os.path.dirname(__file__), "linux_i686_release.json")
        with open(data_file) as data:
            return typing.cast(GitHubReleaseData, json.load(data))

    latest_release_url = urllib.parse.urljoin(GITHUB_API_RELEASES_URL, "latest")
    try:
        with urlopen(latest_release_url) as response:
            release_data = typing.cast(GitHubReleaseData, json.load(response))

    except urllib.error.URLError:
        release_data = fallback_release_data()

    return release_data


def list_pythons() -> dict[str, str]:
    """Returns available python versions for your machine and their download links."""
    system, machine = platform.system(), platform.machine()
    download_link_suffixes = MACHINE_SUFFIX[system][machine]
    # linux suffixes are nested under glibc or musl builds
    if system == "Linux":
        # fallback to musl if libc version is not found
        libc_version = platform.libc_ver()[0] or "musl"
        download_link_suffixes = download_link_suffixes[libc_version]

    is_linux_i686 = system == "Linux" and machine == "i686"
    releases = get_latest_python_releases(is_linux_i686)
    python_releases = [asset["browser_download_url"] for asset in releases["assets"]]

    available_python_links = [
        link
        # Suffixes are in order of preference.
        for download_link_suffix in download_link_suffixes
        for link in python_releases
        if link.endswith(download_link_suffix)
    ]

    python_versions: dict[str, str] = {}
    for link in available_python_links:
        match = PYTHON_VERSION_REGEX.search(link)
        assert match is not None
        python_version = match[1]
        # Don't override already found versions, as they are in order of preference
        if python_version in python_versions:
            continue

        python_versions[python_version] = link

    sorted_python_versions = {
        version: python_versions[version]
        for version in sorted(
            python_versions,
            # sort by semver
            key=lambda version: [int(k) for k in version.split(".")],
            reverse=True,
        )
    }
    return sorted_python_versions


def _parse_python_version(version: str) -> tuple[int, ...]:
    return tuple(int(k) for k in version.split("."))


def resolve_python_version(requested_version: str | None) -> tuple[str, str]:
    pythons = list_pythons()

    if requested_version is None:
        sorted_pythons = sorted(
            pythons.items(),
            key=lambda version_link: _parse_python_version(version_link[0]),
            reverse=True,
        )
        latest_version, download_link = sorted_pythons[0]
        return latest_version, download_link

    for version, version_download_link in pythons.items():
        if version.startswith(requested_version):
            python_version = version
            download_link = version_download_link
            break
    else:
        raise NotAvailable

    return python_version, download_link
